part of '../flutterx_live_data.dart';

/// LiveMap of LiveData
typedef KeyedLiveData<K, V> = LiveMap<K, LiveData<V>>;

/// A [Widget] whose content is built from the observed combination of [KeyedLiveData] and a [LiveData] key
class KeyedLiveDataBuilder<K, V> extends StatelessWidget {
  /// Data list which is observed
  final LiveData<K> dataKey;

  /// Data list which is observed
  final KeyedLiveData<K, V> data;

  /// Fallback LiveData when [data] does not contains [dataKey] if null an exception is thrown
  final LiveData<V>? fallback;

  /// Widget builder function
  final LiveWidgetBuilder<V> builder;

  /// Displayed when [data] is uninitialized, defaults to [SizedBox.shrink]
  final Widget? placeholder;

  /// Creates [builder] from [data]
  const KeyedLiveDataBuilder({
    Key? key,
    required this.dataKey,
    required this.data,
    this.fallback = const LateImmutableLiveData(),
    required this.builder,
    this.placeholder,
  }) : super(key: key);

  @override
  StatelessElement createElement() => _KeyedLiveDataBuilderElement<K, V>(this);

  @override
  Widget build(BuildContext context) {
    final data = (context as _KeyedLiveDataBuilderElement<K, V>)._mediator;
    return data.initialized ? builder(context, data._data) : (placeholder ?? const SizedBox.shrink());
  }

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) => super.debugFillProperties(properties
    ..add(DiagnosticsProperty<LiveData<K>>('dataKey', dataKey))
    ..add(DiagnosticsProperty<KeyedLiveData<K, V>>('data', data))
    ..add(DiagnosticsProperty<LiveData<V>>('fallback', fallback))
    ..add(ObjectFlagProperty<LiveWidgetBuilder<V>>.has('builder', builder)));
}

class _KeyedLiveDataBuilderElement<K, V> extends StatelessElement
    with _LiveDataBuilderMixin<KeyedLiveDataBuilder<K, V>> {
  final MediatorLiveData<V> _mediator = MediatorLiveData<V>(reactive: false);
  Disposable? _observer;

  _KeyedLiveDataBuilderElement(KeyedLiveDataBuilder<K, V> widget) : super(widget);

  @override
  void mount(Element? parent, dynamic newSlot) {
    super.mount(parent, newSlot);
    _mediator.observeForever(onChanged, dispatcher: Dispatcher.postFrame);
  }

  @override
  void unmount() {
    _mediator.removeObserver(onChanged);
    super.unmount();
  }

  @override
  bool dataChanged(KeyedLiveDataBuilder<K, V> oldWidget, KeyedLiveDataBuilder<K, V> newWidget) =>
      oldWidget.dataKey != newWidget.dataKey || oldWidget.data != newWidget.data;

  @override
  void link(KeyedLiveDataBuilder<K, V> widget) => _observer = _mediator.addLiveTransformSource2<K, Map<K, LiveData<V>>>(
      widget.dataKey,
      widget.data,
      (key, data) =>
          data[key] ??
          widget.fallback ??
          (() => throw StateError('data does not contains key $key and no fallback was provider'))());

  @override
  void unlink(KeyedLiveDataBuilder<K, V> widget) {
    _observer?.dispose();
    _observer = null;
  }
}
